//====================================================================================//
//                                                                                    //
//                                    PEPS Main                                       //
//                                                                                    //
//====================================================================================//
//  This File:   peps.C                     Language: C++  (xlC and CC)               //
//  Software:    PEPS 2003                   Advisor: Brigitte.Plateau@imag.fr        //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 13/feb/03                        by: Anne.Benoit@imag.fr             //
//====================================================================================//
#include "peps.h"

peps_name filename;

//====================================================================================//
void Compare_vct()
//====================================================================================//
{
 peps_name name;
 peps_name v_name;
 VCT       vec;
 rp        abs_err, rel_err;
 rp        big, sml;
 int       bpos, spos;
 rp        acc;
 

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 if (Provide_VCT()) {
   VCT::vct.name(name);
   cout << "\nCompare vector '" << name
        << ".vct' (" << VCT::vct.size() << "/"
	<< VCT::vct.real_size() << " states) with ";
   Ask_an_Existing_File_Name(v_name, vct_file);
   vec.read(v_name);
   if (vec.size() != VCT::vct.size())
     cout << "The vectors (" << name << ".vct and " << v_name
          << ".vct) do not have the same size!\n";
   else {
     abs_err = VCT::vct.compare(vec, abs_ind_err);
     VCT::vct.info(acc, big, bpos, sml, spos);
     rel_err = abs_err / sml;
     cout <<   "The maximum absolute error is: " << abs_err
          << "\nThe maximum relative error is: " << rel_err
          << "\nVector '" << name << ".vct' reachable states probability"
          << "\n   max.:  " << big << "   min.: " << sml;
     vec.info(acc, big, bpos, sml, spos);
     cout << "\nVector '" << v_name << ".vct' reachable states probability"
          << "\n   max.:  " << big << "   min.: " << sml << "\n";
   }
 }
}

//====================================================================================//
void Compute_Residue_vct()
//====================================================================================//
{
 peps_name v_name, d_name, vr_name;
 VCT       buf;
 rp        acc, big, sml;
 int       bpos, spos;

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 if (Provide_VCT_CND()) {
   VCT::vct.name(v_name);
   CND::cnd.name(d_name);
   if (VCT::vct.size() != CND::cnd.ss())
     cout << "\nThe vector " << v_name << ".vct and the descritpor "
          << d_name << ".cnd does not have the same size\n";
   else {
     cout << "\nComputing the residual of the solution '" << v_name
          << ".vct' (" << RSS::rss.reach_size() << "/" << VCT::vct.real_size()
          << " states) for the descriptor "
          << d_name << ".cnd (" << CND::cnd.aut() << " automata)\n"
          << "Please enter a probability vector name in which to store the residual\n";
     Ask_a_File_Name(vr_name, vct_file);
     buf.create(VCT::vct.size(), VCT::vct.real_size());
     buf.get_CND();
     buf.alloc_static();
     buf.baptise(vr_name);
     buf.mult_CND(VCT::vct);
     buf.write();
     buf.info(acc, big, bpos, sml, spos);
     if (absv(acc) > round_zero)
       cout << "The result (" << vr_name << ".vct) is not a residual vector\n"
            << "maybe " << d_name << ".cnd is not an infinitesimal generator\n"
            << "maybe " << v_name << ".vct is not a probability vector\n"
            << "or a program error occured!\n";
     else {
       cout << "The largest element in residual vector is " << big << " (" << bpos
            << "),\n" << "therefore the vector " << v_name << ".vct";
       if (big > PRF::prf.min_err())  cout << " is not ";
       else                           cout << " is ";
       cout << "the approximated solution\n"
            << "to the descritor " << d_name << ".cnd (smallest: " << sml
            << " (" << spos << ")\n";
     }
     buf.erase();
     buf.desalloc_static();
   }
 }
}

//====================================================================================//
void Show_Characteristics_vct()
//====================================================================================//
{
 peps_name name;
 int first, size, i;
 rp  sum, big, sml;
 int big_w, sml_w;

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 Provide_VCT();
 first = 0;
 size  = VCT::vct.size();
 VCT::vct.name(name);
 cout << "\nVector '" << name << ".vct' (" << RSS::rss.reach_size()
      << "/" << VCT::vct.size() << " states)";
 VCT::vct.info(sum, big, big_w, sml, sml_w);
 cout << "\nSum of all " << size << " elements: " << sum
      << "\n  - Largest  reachable element (" << big_w << "): " << big
      << "\n  - Smallest reachable element (" << sml_w << "): " << sml
      << "\nFirst elements: ";
 for (i=first; i<first+4; i++)
   cout << VCT::vct.get(i) << " ";
 cout << "\nLast elements: ";
 for (i=size-4; i<size; i++)
   cout << VCT::vct.get(i) << " ";
 cout << "\n";
}

//====================================================================================//
void Analyse_Reachable_States()
//====================================================================================//
{
 peps_name name, output_name;
 ofstream  fout;
 int       first, last, i;
 aut_id    a;
 aut_st    global_state, sizes;
 bool      see_one;

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 if (Provide_VCT_CND()) {
   VCT::vct.name(name);
   cout << "\nVector '" << name
        << ".vct' (" << RSS::rss.reach_size() << "/" << VCT::vct.real_size()
	<< " states)\n\n";
   global_state.create(CND::cnd.aut());
   CND::cnd.get_s(sizes);
   do { first = 0;
        last  = VCT::vct.real_size()-1;
        see_one = false;
        Ask_a_Vector_Range(first, last);
        global_state.st_rank(first, sizes);
        for (i=first; i<last+1; i++,global_state.next(sizes)) {
    	  if (RSS::rss.has(i)) {
            see_one = true;
            for (a=0; a<global_state.size(); a++)
              cout << global_state.st(a) << " ";
            cout << " : " << VCT::vct.get(i) << " (" << i << ")\n";
	      }
        }
        if (! see_one)
          cout << "No reachable state from state " << first << " to " << last << "\n";
        if (Ask_an_Answer("Do you wish to save ALL reachable states in a file")) {
          Ask_a_File_Name(output_name, rea_file);
          if (Permission_to_Proceed(output_name, rea_file)) {
            Open_File(output_name, rea_file, fout);
            fout << "\nVector '" << name << ".vct' ("
                 << RSS::rss.reach_size() << "/" << VCT::vct.real_size() << " states)\n\n";
            global_state.reset();
            for (i=0; i<VCT::vct.real_size(); i++,global_state.next(sizes))
              if (RSS::rss.has(i)) {
                for (a=0; a<global_state.size(); a++)
                  fout << global_state.st(a) << " ";
                fout << " : " << VCT::vct.get(i) << " (" << i << ")\n";
              }
            Close_File(output_name, rea_file, fout);
          }
        }
    } while (Ask_an_Answer("Do you wish to see (on screen) more elements"));
 }
}

//====================================================================================//
void Integrate_Local_States()
//====================================================================================//
{
 peps_name name_vec, name_cnd, name_aut, name_st;
 int       product_size, remaining_size;
 int       i,            k;
 aut_st    state;
 aut_st    sizes;
 aut_id    a;
 st_id     s, first, last;
 rp        sum, acc;
 bool      accumulate;

 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 if (Provide_VCT_CND()) {
   VCT::vct.name(name_vec);
   CND::cnd.name(name_cnd);
   // Verify integrity cnd vct rss ftb (same size)
   FTB::ftb.get_domain(sizes);
   state.create(FTB::ftb.aut());
   product_size   = FTB::ftb.ss();
   remaining_size = VCT::vct.real_size();
   cout << "\nIntegration of the vector '" << name_vec << ".vct' ("
        << remaining_size << " states) solution of '" << name_cnd << ".cnd' ("
        << product_size   << " states)\n";
   Ask_an_Automaton(a);
   Ask_Local_States(a, sizes.sz(a), s);
   cout << "\n";
   if (s != no_st) { first = s;
                     last  = s+1; }
   else {            accumulate = true;
                     acc        = zero;
                     first = 0;
                     last  = sizes.sz(a); }
   for (s=first; s<last; s++) {
     sum = 0;
     state.reset();
     for (i=0,k=0; i<product_size; i++,state.next(sizes))
       if (RSS::rss.remains(i)) {
         if (state.st(a) == s)
           sum += VCT::vct.get(k);
           k++;
       }
     DCT::dct.s_name(a, s, name_st);
     cout << "   " << sum << " for local state '" << name_st << "'\n";
     acc += sum;
   }
   DCT::dct.a_name(a, name_aut);
   if (accumulate)
     cout << " --------------------------\n"
          << "   " << acc << " (total for automaton '" << name_aut << "')\n";
 }
}

//====================================================================================//
void Integrate_Results(const bool write_log)
//====================================================================================//
{
 peps_name name_vec, name_cnd, name_inf, name_func;
 func_id   f;
 int       product_size, remaining_size;
 int       i, k, j;
 aut_st    state;
 aut_st    sizes;
 rp        sum, val;
 
 cout.setf(ios::scientific);
 cout.precision(peps_precision);
 if (Provide_VCT_CND()) {
   VCT::vct.name(name_vec);
   CND::cnd.name(name_cnd);
   INF::inf.name(name_inf);
   // Verify integrity cnd vct rss ftb inf (same size)
   INF::inf.get_domain(sizes);
   state.create(FTB::ftb.aut());
   product_size   = INF::inf.ss();
   remaining_size = VCT::vct.real_size();
   cout << "\nIntegration (" << name_inf << ".inf) of the vector '"
        << name_vec << ".vct' (" << remaining_size << " states) solution of '"
        << name_cnd << ".cnd' (" << product_size   << " states)\n";

   for (f=0; f<INF::inf.size(); f++) {
     sum = 0;
     state.reset();
     for (i=0,j=0,k=0; i<product_size; i++, state.next(sizes))
  	 if (RSS::rss.remains(i)) {
	     if (RSS::rss.has(i)) {   // If the state is reachable, we take the value
  	     val = VCT::vct.get(j); // in VCT::vct.vec
	       j++;                   // We go to the next element of vec
	       sum += val * INF::inf.eval(f, state); // Compute the result
  	   }
	     k++;  // go to the next state
	   }
     DCT::dct.sf_name(f, name_func);
     cout << "Integration of function " << name_func << " : " << sum << "\n";
     if (write_log)
       Log_Out_a_Solution_Function(f, sum);
   }
 }
}

//====================================================================================//
bool Start_Up_CND(VCT & buf)
//====================================================================================//
{
 bool      answer = false;
 peps_name name, n;
 v_typ     initial_vector_type = PRF::prf.i_vec();


 if (!Provide_CND(filename))
   exit(0);
 CND::cnd.name(name);
 cout << "\nSolution of the model '" << name << ".cnd' ("
      << CND::cnd.aut() << " automata - " << RSS::rss.reach_size()
      << "/" << RSS::rss.size() << " states)\n";
 Ask_a_File_Name(name, vct_file);
 VCT::vct.baptise(name);
 if (Permission_to_Proceed(name, vct_file)) {
   VCT::vct.create(RSS::rss.reach_size(), RSS::rss.size());
   VCT::vct.get_CND();
   VCT::vct.baptise(name);
   strcpy(n,name);
  
   if (initial_vector_type != in_vec) {
   	 buf.create(RSS::rss.reach_size(), CND::cnd.ss());
     buf.init(initial_vector_type);
   }
   else
     do {
       Ask_an_Existing_File_Name(name, vct_file);
       buf.read(name);
     } while (buf.real_size() != VCT::vct.real_size());
  
   answer = true;
   buf.baptise(n);
 }
 
 return(answer);
}


//====================================================================================//
//====================================================================================//

//====================================================================================//
//                                  SOLVE_EX Main                                     //
//====================================================================================//
int main(int argc, char *argv[])
{
 VCT       buf;

 cout.setf(ios::scientific);
 cout.precision(16);
 set_new_handler(Communicate_Out_of_Memory);
 Create_Temporary_Directories();
 int solve_op = nothing;
 int integr_op = false;
 peps_name aux;
 
 strcpy(filename,""); 

 if (argc >= 2){
   for (int i=1; i<argc; i++){
     if (argv[i][0] == '-'){
       if (!(strcmp(argv[i], "-sol_meth"))){
         if (solve_op == nothing){
           if (!(strcmp(argv[i+1], "power"))) {
             solve_op = power_std;   i++;
             cout << "Solution method: Power method.\n"; 
           }
           else{
             if (!(strcmp(argv[i+1], "unifor"))){
               solve_op = unifor_std;   i++;                
               cout << "Solution method: Uniformization method.\n"; 
             }
             else{
               cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] << endl; 
               exit (0);
             }
           } 
         }else
           cout << "warning: solution method is already set. Ignoring option \"-sol_meth\"." << endl;
       }// end sol_meth
       if (!(strcmp(argv[i], "-int_func"))){
         integr_op = true;
         cout << "Function integration: true\n";
       }// end int_func
       if (!(strcmp(argv[i], "-precond"))){
         PRF::prf.p_type(diag);
         cout << "Diagonal preconditionning: true\n";
       }// end precond
       if (!(strcmp(argv[i], "-iter"))){
         PRF::prf.max_it(atol(argv[i+1]));   i++;
         cout << "Maximum number of iterations: " << atol(argv[i]) << endl;
       }// end iter 
       if (!(strcmp(argv[i], "-min_err"))){
         PRF::prf.min_err(atof(argv[i+1]));   i++;
         cout << "Tolerance accepted: " << atof(argv[i]) << endl;
       }
       if (!(strcmp(argv[i], "-iter_type"))){
         if (!(strcmp(argv[i+1], "fix"))) {
           PRF::prf.it_type(fix_it); i++;
         }
         else{
           if (!(strcmp(argv[i+1], "stb"))) {
             PRF::prf.it_type(stb_it);  i++;
           }
           else{
             if (!(strcmp(argv[i+1], "cnv"))) {
               PRF::prf.it_type(cnv_it);   i++;
             }
             else{
               cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] << endl;
               exit (0);
             }
           }
         } 
         PRF::prf.sit_type(aux);
         cout << "Stop iteration criterion: " << aux << endl; 
       }
       if (!(strcmp(argv[i], "-err_type"))){
         if (!(strcmp(argv[i+1], "abs_ind"))) {
           PRF::prf.err_type(abs_ind_err);  i++;
         }
         else{
           if (!(strcmp(argv[i+1], "abs_acc"))) {
             PRF::prf.err_type(abs_acc_err);  i++;
           }
           else{
             if (!(strcmp(argv[i+1], "abs_ind"))) {
               PRF::prf.err_type(rel_ind_err);  i++;
             }
             else{
               cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] <<  endl; 
               exit (0);
             }
           }
         } 
         PRF::prf.serr_type(aux);
         cout << "Error Test: " << aux << endl;
       }       
       if (!(strcmp(argv[i], "-vec_type"))){
         if (!(strcmp(argv[i+1], "eq"))) {
           PRF::prf.i_vec(eq_vec);  i++;
         }
         else{
           if (!(strcmp(argv[i+1], "in"))) {
             PRF::prf.i_vec(in_vec);  i++;
           }
           else{
             if (!(strcmp(argv[i+1], "ap"))) {
               PRF::prf.i_vec(ap_vec);  i++;
             }
             else{
               cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] <<  endl; 
               exit (0);
             }
           }
         } 
         PRF::prf.si_vec(aux);
         cout << "Initial vector: " << aux << endl;
       }
       if (!(strcmp(argv[i], "-meth_type"))){
         if (!(strcmp(argv[i+1], "A"))) {
           PRF::prf.method(meth_A); i++;
         }
         else{
           if (!(strcmp(argv[i+1], "B"))) {
             PRF::prf.method(meth_B); i++;
           }
           else{
             if (!(strcmp(argv[i+1], "C"))) {
               PRF::prf.method(meth_C); i++;
             }
             else{
               if (!(strcmp(argv[i+1], "M"))) {
                 PRF::prf.method(meth_M); i++;
               }
               else{
                 cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] <<  endl; 
                 exit (0);
               }
             }
           }
         } 
         PRF::prf.smethod(aux);
         cout << "Vector-Descriptor product method: " << aux << endl;
       }     
       if (!(strcmp(argv[i], "-zin_impl"))){
         if (!(strcmp(argv[i+1], "full"))) {
           PRF::prf.z_impl(zin_full); i++;
         }
         else{
           if (!(strcmp(argv[i+1], "sparse"))) {
             PRF::prf.z_impl(zin_sparse); i++;
           }
           else{
             cout << "error: " << argv[i+1] << " is a invalid option to " << argv[i] << endl; 
             exit (0);
           }
         } 
         PRF::prf.sz_impl(aux);
         cout << "Vector implementation:" << aux << endl;
       }       
//       if ((strcmp(argv[i], "-sol_meth")) && (strcmp(argv[i], "-int_func"))  && (strcmp(argv[i], "-precond"))    && (strcmp(argv[i], "-iter")) && 
//           (strcmp(argv[i], "-min_err"))  && (strcmp(argv[i], "-iter_type")) && (strcmp(argv[i], "-err_type")) && (strcmp(argv[i], "-zin_impl")) &&
//           (strcmp(argv[i], "-vec_type")) && (strcmp(argv[i], "-meth_type"))) {
//         cout << "error: " << argv[i] << " is a invalid option." << endl; 
//         exit (0);
//       }
     }
     else{ //filename
       if (!strcmp(filename, ""))
         strcpy(filename, argv[i]);
       else
        cout << "warning: model name is already set as " << filename << ". Ignoring " << argv[i] << endl; 
     }
   } 
 }

 
 switch (solve_op) {
//====================================================================================//
//                                   Solve Operations                                 //
//====================================================================================//
   case nothing      :   
   case power_std    : if (Start_Up_CND(buf))      Power_CND(buf);
                       if (integr_op)              Integrate_Results(true); 
                       break;

   case unifor_std   : if (Start_Up_CND(buf))      Uniformization_CND(buf, 
                                                     Ask_Time_Transient_Solution());
                       if (integr_op)              Integrate_Results(true);
                       break;
//====================================================================================//
//                                   Vector Operations                                //
//====================================================================================//
   case compare_vct       : Compare_vct();                         break;
   case residue_vct       : Compute_Residue_vct();                 break;
   case charac_vct        : Show_Characteristics_vct();            break;
   case analyse_reachable : Analyse_Reachable_States();            break;
   case integrate_local   : Integrate_Local_States();              break;
   case integrate_results : Integrate_Results(true);               break;

//====================================================================================//
//                                   Other Operations                                 //
//====================================================================================//
   case credits     : Show_Credits();                              break;
   default          : Programming_Error("invalid PEPS option", 0000)
 }
 Say_Goodbye();
 return(0);
}
