//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct2.C                     Language: C++  (xlC and CC)               //
//  Software:    PEPS 2002                   Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2400                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 13/feb/02                        by:     Anne.Benoit@imag.fr         //
//====================================================================================//
#include "peps.h"

      //--------------------------------------------------------------------------//
      //      Static variables to the Internal Functions of Multiply by ...       //
      //                                                                          //
      //   The following five variables are defined as static to reduce the cost  //
      // for allocation and release memory structures in the most sensible point  //
      // of the PEPS solvers.                                                     //
      //--------------------------------------------------------------------------//
VCT    VCT::t;       // temporary vector to store the intermediary results
rp   * VCT::z_in;    // input "slice" (a pointer to an array of vectors elements)
rp   * VCT::z_out;   // output "slice" (a pointer to an array of vectors elements)
aut_st VCT::g_st;    // a global state according to the current descriptor size


vector<gs_pair> VCT::zin;          // input "slice" in sparse format
vector<gs_pair> * VCT::zins;       // array of input  "slice" in sparse format
vector<gs_pair>  VCT::zouts;       // output "slice" in sparse format
vector<info_pair> VCT::info2_zin;  // some informations about the zin
vector<info_pair> VCT::info2_zout; // some informations about the zout

bool * VCT::used_z;                // the used zin or zout
rp ** VCT::zinf;
int * VCT::indice_r;
int ** VCT::useful_f;
int ** VCT::place;

     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method A  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_A(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2403)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2404)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions
 int k2;
 
 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v); // The RSS is the same
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : mfip(CND::cnd.local_mat[a], CND::cnd.state_map, v, a);
   }
 
 for (e=0; e<CND::cnd.events; e++) {
   a = CND::cnd.s_D_order[e].aut(0);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : for (k=0; k<dimension; k++) {
                  	 t.vec[k] = v.vec[k];
                  	 t.positions[k] = v.positions[k];
                     }
                     break;
     case null_mat : t.positions[0] = -1;
                     break;
     case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                     break;
     case func_mat : t.mfis(CND::cnd.synch_mat[e][a],CND::cnd.state_map, v, a);
   }
   
   for (i=1; i<CND::cnd.automata-1; i++) {
     a = CND::cnd.s_D_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : t.positions[0] = -1;
                       i=CND::cnd.automata-1;   // stops the loop
                       break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                       break;
       case func_mat :   t.mfis(CND::cnd.synch_mat[e][a],CND::cnd.state_map, t, a);
     }
   }
   
   a = CND::cnd.s_D_order[e].aut(CND::cnd.automata-1);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : k2=0;
                     for (k=0; k<dimension; k++) {
                  	 if (t.positions[k2]!=-1) {
                  	   if (t.positions[k2]==positions[k]) {
                  	     vec[k] += t.vec[k2];
                  	     k2++;
                  	   }
                  	 }
                  	 else
                  	   break;
                     }
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.synch_mat[e][a], t, a);
                     break;
     case func_mat : mfip(CND::cnd.synch_mat[e][a],CND::cnd.state_map, t, a);
   }
 }
}

     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method B  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_B(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2405)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2406)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions
 bool b = true;

 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v); // The RSS is the same
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : t.becomes(v, CND::cnd.l_B_order[a], CND::cnd.l_B_jumps[a]);
                     t.mfos(CND::cnd.local_mat[a], CND::cnd.state_map, t, a, true);
                     (*this) += t;
   }
 for (e=0; e<CND::cnd.events; e++) {
   t.becomes(v, CND::cnd.s_B_order[e], CND::cnd.s_B_jumps[e]);
   for (i=0; i<CND::cnd.automata; i++) {
     a = CND::cnd.s_D_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : t.positions[0] = -1;
                  	 i=CND::cnd.automata-1;     // stops the loop
                  	 break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
	                   b = false;   // we may have change the positions of t.
                  	 break;
       case func_mat : t.mfos(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a, b);
     }
   }
   (*this) += t;
 }
}

     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method C  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_C(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2407)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2408)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions
 bool b = true;

 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v); // The RSS is the same
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : t.becomes(v, CND::cnd.l_C_order[a], CND::cnd.l_C_jumps[a]);
                     t.mnps(CND::cnd.local_mat[a], CND::cnd.state_map, t, a,
                            CND::cnd.l_C_left[a], CND::cnd.l_C_right[a],
                            CND::cnd.l_C_fst[a], true);
                     (*this) += t;
   }
 for (e=0; e<CND::cnd.events; e++) {
   //------------------------------
   t.get_CND();  // it copies the permutation info of the original vector ('v')
   //------------------------------
   a = CND::cnd.s_D_order[e].aut(0);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : for (k=0; k<dimension; k++) {
                	   t.vec[k] = v.vec[k];
                	   t.positions[k] = v.positions[k];
	                 }
                     break;
     case null_mat : t.positions[0] = -1;
                     break;
     case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                     b = false;  // We may have change the positions of t
                     break;
     case func_mat : t.becomes(v, CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                     t.mnps(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a,
                  	      CND::cnd.s_C_left[e][a],  CND::cnd.s_C_right[e][a],
                  	      CND::cnd.s_C_fst[e][a],   true);
   }
   for (i=1; i<CND::cnd.automata; i++) {
     a = CND::cnd.s_D_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : t.positions[0] = -1;
                  	 i=CND::cnd.automata-1;   // stops the loop
                  	 break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                  	 b = false;  // We may have change the positions of t
                  	 break;
       case func_mat : t.permut(CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                  	 t.mnps(CND::cnd.synch_mat[e][a],   CND::cnd.state_map, t, a,
                      		CND::cnd.s_C_left[e][a],	CND::cnd.s_C_right[e][a],
                      		CND::cnd.s_C_fst[e][a], 	b);
     }
   }
   (*this) += t;
 }
}
 
     //--------------------------------------------------------------------------//
     //     Multiply by a CND  -  Mixed Method  -  'this' = 'v' * 'CND::cnd'     //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_M(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2409)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2410)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions
 int k2;
 bool b = true;


 for (a=0; a<CND::cnd.automata; a++) {
   switch (CND::cnd.l_type[a]) {
     case idnt_term : for (k=0; k<dimension; k++)
                        vec[k] += one;
                      break;
     case null_term : break;
     case cnst_term : mcp(CND::cnd.local_mat[a], v, a);
                      break;
     case A_term    : mfip(CND::cnd.local_mat[a], CND::cnd.state_map, v, a);
                      break;
     case B_term    : t.becomes(v, CND::cnd.l_B_order[a], CND::cnd.l_B_jumps[a]);
                      t.mfos(CND::cnd.local_mat[a], CND::cnd.state_map, t, a, true);
                      (*this) += t;
                      break;
     case C_term    : t.becomes(v, CND::cnd.l_C_order[a], CND::cnd.l_C_jumps[a]);
                      t.mnps(CND::cnd.local_mat[a], CND::cnd.state_map, t, a,
                   	      CND::cnd.l_C_left[a],  CND::cnd.l_C_right[a],
                   	      CND::cnd.l_C_fst[a],   true);
                      (*this) += t;
   }
 }

 for (e=0; e<CND::cnd.events; e++) {
   //------------------------------
   t.get_CND();  // it copies the permutation info of the original vector ('v')
   //------------------------------
   switch (CND::cnd.s_type[e]) {
     case idnt_term : for (k=0; k<dimension; k++)
                        vec[k] += one;
                      break;
     case null_term : break;
     case cnst_term : 
     case A_term    : a = CND::cnd.s_D_order[e].aut(0);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : for (k=0; k<dimension; k++) {
                                   	   t.vec[k] = v.vec[k];
                                   	   t.positions[k] = v.positions[k];
                                   	 }
                                        break;
                        case null_mat : t.positions[0] = -1;
                                        break;
                        case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                                        break;
                        case func_mat : t.mfis(CND::cnd.synch_mat[e][a],CND::cnd.state_map, v, a);
                      }
                      for (i=1; i<CND::cnd.automata-1; i++) {
                        a = CND::cnd.s_D_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : t.positions[0] = -1;
                                     	 i=CND::cnd.automata-1;   // stops the loop
                                     	 break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                     	 break;
                          case func_mat : t.mfis(CND::cnd.synch_mat[e][a],CND::cnd.state_map, t, a);
                        }
                      }
                      a = CND::cnd.s_D_order[e].aut(CND::cnd.automata-1);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : k2=0;
                                   	 for (k=0; k<dimension; k++) {
                                   	   if (t.positions[k2]!=-1) {
                                   	     if (t.positions[k2]==positions[k]) {
                                   	       vec[k] += t.vec[k2];
                                   	       k2++;
                                   	     }
                                   	   }
                                   	   else
                                   	     k=dimension;
                                   	 }
                                   	 break;
                        case null_mat : break;
                        case cnst_mat : mcp(CND::cnd.synch_mat[e][a], t, a);
                                        break;
                        case func_mat : mfip(CND::cnd.synch_mat[e][a],CND::cnd.state_map, t, a);
                      }
                      break;
     case B_term    : t.becomes(v, CND::cnd.s_B_order[e], CND::cnd.s_B_jumps[e]);
                      for (i=0; i<CND::cnd.automata; i++) {
                        a = CND::cnd.s_D_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : t.positions[0] = -1;
                                     	 i=CND::cnd.automata-1;     // stops the loop
                                     	 break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                     	 b = false; // We may have change the positions of t
                                     	 break;
                          case func_mat : t.mfos(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a, b);
                        }
                      }
                      (*this) += t;
                      break;
     case C_term    : a = CND::cnd.s_D_order[e].aut(0);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : for (k=0; k<dimension; k++) {
                                   	   t.vec[k] = v.vec[k];
                                   	   t.positions[k] = v.positions[k];
                                   	 }
                                        break;
                        case null_mat : t.positions[0] = -1;
                                        break;
                        case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                                        b = false;
                                        break;
                        case func_mat : t.becomes(v, CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                                        t.mnps(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a,
                                     	      CND::cnd.s_C_left[e][a],  CND::cnd.s_C_right[e][a],
                                     	      CND::cnd.s_C_fst[e][a],   true);
                      }
                      for (i=1; i<CND::cnd.automata; i++) {
                        a = CND::cnd.s_B_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : t.positions[0] = -1;
                                     	 i=CND::cnd.automata-1;   // stops the loop
                                     	 break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                     	 b = false;
                                     	 break;
                          case func_mat : t.permut(CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                                     	 t.mnps(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a,
                                         		CND::cnd.s_C_left[e][a],  CND::cnd.s_C_right[e][a],
                                         		CND::cnd.s_C_fst[e][a],   b);
                        }
                      }
                      (*this) += t;
   }
 } 
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //     It multiplies a vector 'v' by a normal factor without functions (non //
     // identity matrix in 'm') adding the result to the vector 'this'.          //
     //--------------------------------------------------------------------------//
void VCT::mcp(san_mat & m, const VCT & v, const aut_id a)
{
 int i, ni    = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;              // pointer to pick vector elements for the slices
 int k2, j, j2, rr, dim;
 int jinit, jinit2;
 div_t zin_pos;
 gs_pair tmp;
 int_pair tmp2;
   
 j=0; j2=0;
 dim = v.size();
   
 if (PRF::prf.z_impl() == zin_sparse) { // sparse zin
   info_pair tmpin, tmpout;
   vector<info_pair>::const_iterator izin, izout;
        // iterators to traverse the 'info_zin' and 'info_zout' vectors
   zin.reserve(ni);
   info2_zin.reserve(dim);
   info2_zout.reserve(dimension);
   for (l=0; l<nleft; l++) {                            // for the left sub-space:
     // We create the info_zin array (informations on zin)
     while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
	   tmp2.first=zin_pos.quot;
  	 tmp2.second=j;
	   tmpin.first=zin_pos.rem;
  	 tmpin.second=tmp2;
  	 info2_zin.push_back(tmpin);
  	 j++;
     }
     // We create the info_zout array (informations on zout)
     while ((j2<dimension) && (positions[j2]<(l+1)*njump) && (positions[j2]!=-1)) {
  	 zin_pos = div(positions[j2]-l*njump,nright);
  	 tmp2.first=zin_pos.quot;
  	 tmp2.second=j2;
  	 tmpout.first=zin_pos.rem;
  	 tmpout.second=tmp2;
  	 info2_zout.push_back(tmpout);
  	 j2++;
     }
     // We sort info_zin and info_zout
     sort(info2_zin.begin(), info2_zin.end(), comp2());
     sort(info2_zout.begin(), info2_zout.end(), comp2());
     izin  = info2_zin.begin();
     izout = info2_zout.begin();
     for (r=0;r<nright;r++) {
  	 if ((*izout).first == r) {
	     zin.erase(zin.begin(), zin.end());
  	   while ((*izin).first < r) 
  	     ++izin; // we go to the good zin
  	   while ((izin != info2_zin.end()) && ((*izin).first == r)) {
	       // Extraction of zin
	       tmp2 = (*izin).second;
  	     tmp.first = tmp2.first;
  	     tmp.second = v.vec[tmp2.second];
  	     zin.push_back(tmp);
	       ++izin;
  	   }
	     // sort the current zin
  	   sort(zin.begin(),zin.end(), comp());
  	   // Now that we have the current zin, we compute each element
	     // of the resulting vector, and store the results in vec.
         while ((izout != info2_zout.end()) && ((*izout).first == r)) {
	       tmp2 = (*izout).second;
  	     vec[tmp2.second] += m.sparse_elem_left_mult2(zin, tmp2.first);
	       ++izout;
	     }
	   }
     }
     // Erase everything
     info2_zin.erase(info2_zin.begin(), info2_zin.end());
     info2_zout.erase(info2_zout.begin(), info2_zout.end());
   }
 } // end sparse zin
 else { // zin in full implementation
   for (l=0; l<nleft; l++) {                            // for the left sub-space:
     // Initialisations
     for (r=0; r<nright; r++)
	   for (i=0; i<ni; i++)
  	   zinf[r][i]=0;
     // We create the zinf array for the sparse multiplication
     while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
  	 zinf[zin_pos.rem][zin_pos.quot] = v.vec[j];
  	 j++;
     }
     // Here, we have all the zinf for the l portion ; we can make the 
     // vector-matrix multiplication in sparse : we compute each element
     // of the resulting vector, and store the results in vec.
     while ((j2<dimension) && (positions[j2]<(l+1)*njump)) {
  	 zin_pos = div(positions[j2]-l*njump,nright);
  	 vec[j2] += m.sparse_elem_left_mult(zinf[zin_pos.rem], zin_pos.quot);
  	 j2++;
     }
   } //end of left sub-space    
 } //end of zin in full implementation
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') adding the result to the vector 'this'.          //
     //                                                                          //
     // Obs: The function evaluations are performed inside the inner most loop   //
     // (no permutations are need).                                              //
     //--------------------------------------------------------------------------//
void VCT::mfip(san_mat & m, const st_map * sm, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;               // pointer to pick vector elements for the slices
 int k2, j, j2, rr, dim;
 int jinit, jinit2;
 div_t zin_pos;
 gs_pair tmp;
 int_pair tmp2;

 g_st.reset();                                        // global state reset
 j=0;j2=0;
 dim = v.size();
 if (PRF::prf.z_impl() == zin_sparse) { // sparse zin
   info_pair tmpin, tmpout;
   vector<info_pair>::const_iterator izin, izout;
   // iterators to traverse the 'info_zin' and 'info_zout' vectors
   zin.reserve(ni);
   info2_zin.reserve(dim);
   info2_zout.reserve(dim);
   for (l=0; l<nleft; l++) {                            // for the left sub-space:
     // We create the info_zin array (informations on zin)
     while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
  	 tmp2.first=zin_pos.quot;
	   tmp2.second=j;
  	 tmpin.first=zin_pos.rem;
	   tmpin.second=tmp2;
  	 info2_zin.push_back(tmpin);
  	 j++;
     }
     // We create the info_zout array (informations on zout)
     while ((j2<dimension) && (positions[j2]<(l+1)*njump) && (positions[j2]!=-1)) {
  	 zin_pos = div(positions[j2]-l*njump,nright);
  	 tmp2.first=zin_pos.quot;
  	 tmp2.second=j2;
  	 tmpout.first=zin_pos.rem;
  	 tmpout.second=tmp2;
  	 info2_zout.push_back(tmpout);
  	 j2++;
     }
     // We sort info_zin and info_zout
     sort(info2_zin.begin(), info2_zin.end(), comp2());
     sort(info2_zout.begin(), info2_zout.end(), comp2());
     izin  = info2_zin.begin();
     izout = info2_zout.begin();
     for (r=0;r<nright;r++) {
  	 if ((*izout).first == r) {
	     zin.erase(zin.begin(), zin.end());
  	   while ((*izin).first < r) 
  	     ++izin; // we go to the good zin
  	   while ((izin != info2_zin.end()) && ((*izin).first == r)) {
  	     // Extraction of zin
	       tmp2 = (*izin).second;
  	     tmp.first = tmp2.first;
  	     tmp.second = v.vec[tmp2.second];
  	     zin.push_back(tmp);
  	     ++izin;
	     }
  	   // sort the current zin
  	   sort(zin.begin(),zin.end(), comp());
         // Now that we have the current zin, we compute each element
  	   // of the resulting vector, and store the results in vec.
  	   m.jit_eval(g_st, a, sm);                           // evaluates the matrix
  	   while ((izout != info2_zout.end()) && ((*izout).first == r)) {
  	     tmp2 = (*izout).second;
  	     vec[tmp2.second] += m.sparse_elem_left_mult2(zin, tmp2.first);
  	     ++izout;
	     }
  	 }
  	 g_st.next(v.sizes, a, Right);                       // changes the state right
     }
     // Erase everything
     info2_zin.erase(info2_zin.begin(), info2_zin.end());
     info2_zout.erase(info2_zout.begin(), info2_zout.end());
     g_st.next(v.sizes, a, Left);                       // changes the state left     
   }
 } // end sparse zin
 else { // full zin implementation
   for (l=0; l<nleft; l++) {                        // for the left subspace
     // Initialisations of the array : null
     for (r=0; r<nright; r++) {
   	used_z[r] = false;
  	 indice_r[r] = 0;
  	 for (i=0; i<ni; i++)
	     zinf[r][i] = 0;
     }
     // We create the zinf array for the sparse multiplication
     while ((j<dim) && (v.positions[j]<(l+1)*njump)  && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
  	 rr = zin_pos.rem;
  	 zinf[rr][zin_pos.quot] = v.vec[j];
  	 j++;
     }
     while ((j2<dimension) && (positions[j2]<(l+1)*njump)&& (positions[j2]!=-1)) {
  	 zin_pos = div(positions[j2]-l*njump,nright);
  	 rr = zin_pos.rem;
  	 used_z[rr] = true;
  	 useful_f[rr][indice_r[rr]] = zin_pos.quot;
  	 place[rr][indice_r[rr]] = j2;
  	 indice_r[rr]++;
  	 j2++;
     }
     // Here, we have all the zinf for the l portion ; we can make the 
     // vector-matrix multiplication in sparse : we compute each element
     // of the resulting vector, and store the results in vec.
     for (r=0; r<nright; r++) {                         // for the right sub-space:
  	 if (used_z[r]) {
	     m.jit_eval(g_st, a, sm);                           // evaluates the matrix
  	   for (i=0; i<indice_r[r]; i++)
	       vec[place[r][i]] += m.sparse_elem_left_mult(zinf[r], useful_f[r][i]);
	   }
  	 g_st.next(v.sizes, a, Right);                       // changes the state right
     }     
     g_st.next(v.sizes, a, Left);                       // changes the state left     
   } //end left subspace
   //   delete everything;
   //delete[] used_z;
 } // end full zin
}
     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //     It multiplies a vector 'v' by a normal factor without functions (non //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //--------------------------------------------------------------------------//
void VCT::mcs(san_mat & m, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;                    // pointer to pick vector elements for the slices
 int k2, j, j2, rr, dim;
 int jinit, jinit2;
 div_t zin_pos;
 vector<gs_pair>::const_iterator iter; // an iterator to traverse the 'zouts' vector
 gs_pair tmp;
 int_pair tmp2;
 info_pair tmpin;
 vector<info_pair>::const_iterator izin;
 
 j=0; j2=0; k2=0;
 dim = v.size();
 zouts.reserve(dimension);
 zin.reserve(ni);
 info2_zin.reserve(dim);
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   // We create the info_zin array (informations on zin)
   while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
	 zin_pos = div(v.positions[j]-l*njump,nright);
	 tmp2.first=zin_pos.quot;
	 tmp2.second=j;
	 tmpin.first=zin_pos.rem;
	 tmpin.second=tmp2;
	 info2_zin.push_back(tmpin);
	 j++;
   }
   // We sort info_zin
   sort(info2_zin.begin(), info2_zin.end(), comp2());
   izin  = info2_zin.begin();
   for (r=0;r<nright;r++) {
	 if ((*izin).first == r) {
	   zin.erase(zin.begin(), zin.end());
	   while ((izin != info2_zin.end()) && ((*izin).first == r)) {
	     // Extraction of zin
	     tmp2 = (*izin).second;
	     tmp.first = tmp2.first;
	     tmp.second = v.vec[tmp2.second];
	     zin.push_back(tmp);
	     ++izin;
	   }
	   // sort the current zin
	   sort(zin.begin(),zin.end(), comp());
	   // We have the current zin ; we multiply it with the matrix and
	   // find which element are not nul. Those elements are stored in zouts.
	   m.sparse_left_mult2(zouts, zin, r, nright);
	 }
   }
   // put the zouts results in vec and positions
   sort(zouts.begin(), zouts.end(), comp());
   iter = zouts.begin();
   while ((iter != zouts.end()) && (k2<dimension)) {
	 //cout << "element non nul a la position " << (*iter).first + l*njump << "\n";
	 positions[k2]=(*iter).first + l*njump;
	 vec[k2]=(*iter).second;
	 ++iter; k2++;
   }
   if (k2==dimension)
	 cout << "*** Got out of RSS : you should use full vector implementation ***\n";

   // Erase everything
   info2_zin.erase(info2_zin.begin(), info2_zin.end());
   info2_zout.erase(info2_zout.begin(), info2_zout.end());      
   zouts.erase(zouts.begin(), zouts.end());
 }
 if (k2<dimension)
   positions[k2]=-1;
}
     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The function evaluations are performed inside the inner most loop   //
     // (no permutations are need).                                              //
     //--------------------------------------------------------------------------//
void VCT::mfis(san_mat & m, const st_map * sm, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int pos       = v.p_order.pos(a);     // position of automaton 'a' in the list
 int idx, base = 0;                    // pointer to pick vector elements for the slices
 int k2, j, j2, rr, dim;
 int jinit, jinit2;
 div_t zin_pos;
 vector<gs_pair>::const_iterator iter;  // an iterator to traverse the 'zouts' vector
 gs_pair tmp;
 int_pair tmp2;

 g_st.reset();                                                  // global state reset
 j=0; j2=0; k2=0;
 dim = v.size();
 zouts.reserve(dimension);

 info_pair tmpin;
 vector<info_pair>::const_iterator izin;
        // iterators to traverse the 'info_zin' vector
 zin.reserve(ni);
 info2_zin.reserve(dim);
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   // We create the info_zin array (informations on zin)
   while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
	 zin_pos = div(v.positions[j]-l*njump,nright);
	 tmp2.first=zin_pos.quot;
	 tmp2.second=j;
	 tmpin.first=zin_pos.rem;
	 tmpin.second=tmp2;
	 info2_zin.push_back(tmpin);
	 j++;
   }
   // We sort info_zin
   sort(info2_zin.begin(), info2_zin.end(), comp2());
   izin  = info2_zin.begin();
   for (r=0;r<nright;r++) {
	 if ((*izin).first == r) {
	   zin.erase(zin.begin(), zin.end());
	   while ((izin != info2_zin.end()) && ((*izin).first == r)) {
	     // Extraction of zin
	     tmp2 = (*izin).second;
	     tmp.first = tmp2.first;
	     tmp.second = v.vec[tmp2.second];
	     zin.push_back(tmp);
	     ++izin;
	   }
	   // sort the current zin
	   sort(zin.begin(),zin.end(), comp());
  
	   // We have the current zin ; we multiply it with the matrix and
	   // find which element are not nul. Those elements are stored in zouts.
	   m.jit_eval(g_st, a, sm);                           // evaluates the matrix
	   m.sparse_left_mult2(zouts, zin, r, nright);
	 }
	 g_st.next(v.sizes, a, Right);                       // changes the state right
   }
   // put the zouts results in vec and positions
   sort(zouts.begin(), zouts.end(), comp());
   iter = zouts.begin();
      
   while ((iter != zouts.end()) && (k2<dimension)) {
     positions[k2]=(*iter).first + l*njump;
	 vec[k2]=(*iter).second;
	 ++iter; k2++;
   }
   if (k2==dimension)
	 cout << "*** Got out of RSS : you should use full vector implementation ***\n";

   // Erase everything
   info2_zin.erase(info2_zin.begin(), info2_zin.end());
   info2_zout.erase(info2_zout.begin(), info2_zout.end());      
   zouts.erase(zouts.begin(), zouts.end());
   g_st.next(v.sizes, a, Left);                       // changes the state left
 }
 if (k2<dimension)
   positions[k2]=-1;
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The function evaluations are performed outside the inner most loop  //
     // (permutations are need).                                                 //
     //--------------------------------------------------------------------------//
void VCT::mfos(san_mat & m, const st_map * sm, const VCT & v, const aut_id a, const bool b)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int pos       = v.p_order.pos(a);     // position of automaton 'a' in the list
 int idx, base = 0;                    // pointer to pick vector elements for the slices
 int k2, j, j2, rr, dim;
 int jinit, jinit2;
 div_t zin_pos;
 vector<gs_pair>::const_iterator iter;  // an iterator to traverse the 'zouts' vector
 gs_pair tmp;
 int_pair tmp2;

 g_st.reset();                                        // global state reset
 j=0; j2=0;
 dim = v.size();
 if (!b) {
   zouts.reserve(dimension);
   k2=0;
 }

 if ((PRF::prf.z_impl() == zin_sparse)||(!b)) { // sparse zin
   info_pair tmpin, tmpout;
   vector<info_pair>::const_iterator izin, izout;
   // iterators to traverse the 'info_zin' and 'info_zout' vectors
   zin.reserve(ni);
   info2_zin.reserve(dim);
   if (b) 
     info2_zout.reserve(dim);
   for (l=0; l<nleft; l++) {                            // for the left sub-space:
     m.jit_eval(g_st, a, sm);                               // evaluates the matrix
     // We create the info_zin array (informations on zin)
     while ((j<dim) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
  	 tmp2.first=zin_pos.quot;
  	 tmp2.second=j;
  	 tmpin.first=zin_pos.rem;
  	 tmpin.second=tmp2;
  	 info2_zin.push_back(tmpin);
	   j++;
     }
     // We sort info_zin
     sort(info2_zin.begin(), info2_zin.end(), comp2());
     izin  = info2_zin.begin();
     if (b) { // b is true -> take into account the zout positions (that gives
	        // the non nul values
	   // We create the info_zout array (informations on zout)
  	 while ((j2<dimension) && (positions[j2]<(l+1)*njump) && (positions[j2]!=-1)) {
  	   zin_pos = div(positions[j2]-l*njump,nright);
  	   tmp2.first=zin_pos.quot;
  	   tmp2.second=j2;
  	   tmpout.first=zin_pos.rem;
  	   tmpout.second=tmp2;
  	   info2_zout.push_back(tmpout);
  	   j2++;
	   }
  	 // We sort info_zout
  	 sort(info2_zout.begin(), info2_zout.end(), comp2());
  	 izout = info2_zout.begin();
	   for (r=0;r<nright;r++) {
  	   if ((*izout).first == r) {
	       zin.erase(zin.begin(), zin.end());
	       while ((*izin).first < r)
	         ++izin; // we go to the good zin
 	      while ((izin != info2_zin.end()) && ((*izin).first == r)) {
	         // Extraction of zin
	         tmp2 = (*izin).second;
  	       tmp.first = tmp2.first;
  	       tmp.second = v.vec[tmp2.second];
	         zin.push_back(tmp);
	         ++izin;
	       }
	       // sort the current zin
  	     sort(zin.begin(),zin.end(), comp());
           // Now that we have the current zin, we compute each element
	       // of the resulting vector, and store the results in vec.
  	     while ((izout != info2_zout.end()) && ((*izout).first == r)) {
	         tmp2 = (*izout).second;
  	       vec[tmp2.second] = m.sparse_elem_left_mult2(zin, tmp2.first);
	         ++izout;
	       }
	     }
	   }
	   info2_zout.erase(info2_zout.begin(), info2_zout.end());      
     }
	 else {// b is false -> we have to find the non null values
  	 for (r=0;r<nright;r++) {
	     if ((*izin).first == r) {
	       zin.erase(zin.begin(), zin.end());
	       while ((izin != info2_zin.end()) && ((*izin).first == r)) {
  	       // Extraction of zin
	         tmp2 = (*izin).second;
	         tmp.first = tmp2.first;
	         tmp.second = v.vec[tmp2.second];
	         zin.push_back(tmp);
  	       ++izin;
	       }
  	     // sort the current zin
	       sort(zin.begin(),zin.end(), comp());
	       // We have the current zin ; we multiply it with the matrix and
	       // find which element are not nul. Those elements are stored in zouts.
  	     m.sparse_left_mult2(zouts, zin, r, nright);
	     }
	   }
  	 // put the zouts results in vec and positions
	   sort(zouts.begin(), zouts.end(), comp());
  	 iter = zouts.begin();
	   while ((iter != zouts.end()) && (k2<dimension)) {
	     //cout<< "element non nul a la position " <<(*iter).first + l*njump << "\n";
  	   positions[k2]=(*iter).first + l*njump;
	     vec[k2]=(*iter).second;
  	   ++iter; k2++;
	   }
  	 if (k2==dimension)
	     cout<<"*** Got out of RSS : you should use full vector implementation ***\n";
	   zouts.erase(zouts.begin(), zouts.end());
     }
     // Erase everything
     info2_zin.erase(info2_zin.begin(), info2_zin.end());
     g_st.next(v.p_order, v.sizes, pos, Left);          // left-space block
   }
   if ((!b) && (k2<dimension))
     positions[k2]=-1;
 } // end sparse zin
 else { // full zin implementation
   for (l=0; l<nleft; l++) {                        // for the left subspace
     m.jit_eval(g_st, a, sm);                           // evaluates the matrix
     // Initialisations of the array : null
     for (r=0; r<nright; r++) {
	   indice_r[r] = 0;
  	 for (i=0; i<ni; i++)
	     zinf[r][i] = 0;
     }
     // We create the zinf array for the sparse multiplication
     while ((j<dimension) && (v.positions[j]<(l+1)*njump) && (v.positions[j]!=-1)) {
  	 zin_pos = div(v.positions[j]-l*njump,nright);
  	 rr = zin_pos.rem;
  	 zinf[rr][zin_pos.quot] = v.vec[j];
	   j++;
     }
     while ((j2<dimension) && (positions[j2]<(l+1)*njump) && (positions[j2]!=-1)) {
  	 zin_pos = div(positions[j2]-l*njump,nright);
  	 rr = zin_pos.rem;
	   useful_f[rr][indice_r[rr]] = zin_pos.quot;
  	 place[rr][indice_r[rr]] = j2;
	   indice_r[rr]++;
  	 j2++;
     }
     // Here, we have all the zinf for the l portion ; we can make the 
     // vector-matrix multiplication in sparse : we compute each element
     // of the resulting vector, and store the results in vec.
     for (r=0; r<nright; r++)                         // for the right sub-space:
  	 for (i=0; i<indice_r[r]; i++)
	     vec[place[r][i]] = m.sparse_elem_left_mult(zinf[r], useful_f[r][i]);
     g_st.next(v.p_order, v.sizes, pos, Left);          // left-space block
   } //end left subspace
 } // end full zin
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The permutation of the input vector is performed before this        //
     // function is called, the function evaluation is performed outside the     //
     // inner most loop (no permutations are need in this function).             //
     //--------------------------------------------------------------------------//
void VCT::mnps(san_mat & m, const st_map * sm, const VCT & v, const aut_id a,
               const int nleft, const int nright, const aut_id pos, const bool b)
{
 int i, ni = v.sizes.sz(a);    // size of the automaton 'a'
 int l; // nleft               // space size of the parameters of automaton 'a'
 int r; // nright              // space size of the non parameters of automaton 'a'
 int njump = nright * ni;      // 'nright' times 'ni'
        // pos                 // position of the first non parameter of 'a' in the list
 int base = 0;                 // pointer to pick vector elements for the slices
 int dim = v.size();
 int j=0, j2=0, k2=0;
 div_t zin_pos;
 vector<gs_pair>::const_iterator iter; // an iterator to traverse zouts
    
 g_st.reset();                                        // global state reset
 if (!b) 
   zouts.reserve(dimension);
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   m.jit_eval(g_st, a, sm);                               // evaluates the matrix
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     // initialize the z_in
     for (i=0;i<ni;i++)
	   z_in[i] = 0;
     // extract the z_in
     while ((j<dim)&&(v.positions[j]<(base+ni)) && (v.positions[j]!=-1)) {
  	 z_in[v.positions[j]-base] = v.vec[j];
	   j++;
     }
     if (b)       
  	 // vector/matrix multiplication (elt/column)
	   while ((j2<dimension)&&(positions[j2]<(base+ni)) && (positions[j2]!=-1)) {
  	   vec[j2] = m.sparse_elem_left_mult(z_in, positions[j2]-base);
	     j2++;
  	 }
     else 
	   m.sparse_left_mult2(zouts, z_in, base);       
     base += ni;
   } // end of right space
   if (!b) {
     // Get the result of zouts in the vector 'this'
     iter = zouts.begin();
     while ((iter != zouts.end()) && (k2<dimension)) {
  	 positions[k2] = (*iter).first;
	   vec[k2] = (*iter).second;
  	 ++iter; k2++;
     }
     if (k2==dimension)
	   cout << "*** Got out of RSS : you should use full vector implementation ***\n";
   }
   if (!b) 
     zouts.erase(zouts.begin(), zouts.end());     
   g_st.next(v.p_order, v.sizes, pos, Left);          // pass to the next
 }                                                  // left-space block
 if ((!b) && (k2<dimension))
   positions[k2] = -1;
}

      //--------------------------------------------------------------------------//
      //      Multiply by a CND  -  generic switch among methods A, B, C and M    //
      //--------------------------------------------------------------------------//
void VCT::mult_CND(const VCT & v)
{
 int k;

 for (k=0; k<dimension; k++) {                    // dot product of the diagonal
   positions[k] = v.positions[k];
   vec[k] = v.vec[k] * CND::cnd.diagonal[k];
 }

  switch(PRF::prf.method()) {                        // switch methods A, B, C or M
    case meth_A : mult_CND_A(v); break;
    case meth_B : mult_CND_B(v); break;
    case meth_C : mult_CND_C(v); break;
    case meth_M : mult_CND_M(v);
  }
}

     //--------------------------------------------------------------------------//
     //                Multiply by the inverse of the CND diagonal               //
     //--------------------------------------------------------------------------//
void VCT::solv_D()
{
 int k;

 for (k=0; k<dimension; k++)       // dot product of the inverse of
   vec[k] /= CND::cnd.diagonal[positions[k]];                //    the diagonal
}

     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // Diagonal       - uses the diagonal of the 'CND::cnd' descriptor as       //
     //                  preconditionner.                                        //
     //                                          'this = (v * CND) * diag(CND)'  //
     //--------------------------------------------------------------------------//
void VCT::pre_diag(const VCT & v)
{
 mult_CND(v);                          // it performs 'this = v * CND'
 solv_D();                             // it performs 'this = (v * CND) * diag(CND)'
}
     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // Polynomial     - uses the 'CND::cnd' as the preconditionner and          //
     //                  aproximates its inverse by the polynome:                //
     //                  'I + CND + CND^2 + CND^3 + ...'                         //
     //                              'this = v * CND * (I + CND + CND^2 + ...')  //
     //--------------------------------------------------------------------------//
void VCT::pre_plyn(const VCT & v)
{
 int n_terms = PRF::prf.poly_k();
 VCT zz(dimension, real_dim);

 mult_CND(v);                          // it performs 'zz = this = v * CND'
 zz.copy(*this);
 for (int i=1; i<n_terms; i++) {       // for all polynomial terms
   mult_CND(*this);                       // it performs 'this *= CND'
   add(zz);                               // it performs 'this = (v * CND) + this'
 }
}
